﻿Public Class Maze

    Property Grid As Corners(,)

    ' Set to indicate that walls should be shared between adjacent cells
    Property ShareWalls As Boolean = True

    ' Displacement vectors (cells east, north, west, south of the current cell respectively)
    Public Shared Property Directions As New List(Of Point) From {{1, 0}, {0, -1}, {-1, 0}, {0, 1}}

#Region "Constructors"
    Public Sub New(ByVal Grid(,) As Corners)
        Me.Grid = Grid
    End Sub

    Public Sub New(ByVal Columns As Integer, ByVal Rows As Integer)
        SetSize(Columns, Rows)
    End Sub
#End Region
#Region "Dimensions"
    Public ReadOnly Property Dimensions() As Size
        Get
            Return New Size(Columns, Rows)
        End Get
    End Property

    Public ReadOnly Property Columns() As Integer
        Get
            Return Grid.GetLength(0)
        End Get
    End Property

    Public ReadOnly Property Rows() As Integer
        Get
            Return Grid.GetLength(1)
        End Get
    End Property

    Default Public ReadOnly Property Cell(ByVal Coordinates As Point) As Corners
        Get
            Return Grid(Coordinates.X, Coordinates.Y)
        End Get
    End Property

    Default Public ReadOnly Property Cell(ByVal X As Integer, ByVal Y As Integer) As Corners
        Get
            Return Grid(X, Y)
        End Get
    End Property

    Public ReadOnly Property CellOrDefault(ByVal Coordinates As Point) As Corners
        Get
            Static Garbage As New Corners
            Return If(IsCell(Coordinates), Cell(Coordinates), Garbage)
        End Get
    End Property
#End Region

    Public Sub SetSize(ByVal newColumns As Integer, ByVal newRows As Integer)

        ReDim Grid(newColumns - 1, newRows - 1)

        ' Initialize all mazes
        For x = 0 To newColumns - 1
            For y = 0 To newRows - 1
                Grid(x, y) = New Corners With {.Parent = Me, .X = x, .Y = y}
            Next
        Next

    End Sub

    Public Sub Generate(ByVal Start As Point)

        Dim Rnd As New Random, Cells As New Stack(Of Point)()
        Dim Current = Start, Branches() As Adjacent = Nothing

        Do Until Cells.Count = 0 AndAlso Branches IsNot Nothing
            ' Handle this cell
            Cell(Current).Visited = True
            Branches = GetNeighbors(Current).Where(Function(x) Not Cell(x.Coords).Visited).ToArray

            If Branches.Count > 0 Then
                Dim Index = Rnd.Next(0, Branches.Count)
                Cells.Push(Current)
                Cell(Current).Corner(Branches(Index).Direction) = False
                Current = Branches(Index).Coords
            Else
                Current = Cells.Pop
            End If
        Loop

        ClearVisited()

    End Sub

    Public Function FindPath(ByVal Start As Point, ByVal Goal As Point, ByRef Solution As List(Of Point)) As Boolean

        Dim Cells As New Stack(Of Point)()
        Dim Current = Start, Branch As Adjacent

        Do Until Cells.Count = 0 AndAlso Branch IsNot Nothing
            ' See if we've reached the goal
            If Current = Goal Then
                Solution.Add(Goal) : Solution.AddRange(Cells.ToArray)
                FindPath = True : Exit Do
            End If

            ' Handle this cell
            Cell(Current).Visited = True
            Branch = GetNeighbors(Current).Where(
                Function(x) Not Cell(x.Coords).Visited AndAlso Not Cell(Current).Corner(x.Direction)).FirstOrDefault

            If Branch IsNot Nothing Then
                Cells.Push(Current)
                Current = Branch.Coords
            Else
                Current = Cells.Pop
            End If
        Loop

        ClearVisited()

    End Function

    Public Sub Resize(ByVal newColumns As Integer, ByVal newRows As Integer)
        Dim Old = New Maze(Grid)
        SetSize(newColumns, newRows)
        Old.CopyTo(Me)
    End Sub

    Public Function GetNeighbors(ByVal Cell As Point) As IEnumerable(Of Adjacent)
        Return Directions.Select(Function(d, i) New Adjacent() With {.Direction = i, .Coords = Cell + d}).
                          Where(Function(x) IsCell(x.Coords))
    End Function

    Public Function IsCell(ByVal Cell As Point) As Boolean
        Return Cell.X >= 0 AndAlso Cell.Y >= 0 AndAlso Cell.X < Columns AndAlso Cell.Y < Rows
    End Function

    ' Copies one submatrix into another (note, makes reference copies)
    Public Sub CopyTo(ByVal Destination As Maze, ByVal SourceOffset As Point, ByVal DestOffset As Point, ByVal MaximumSize As Size)
        For x = 0 To {Me.Columns - SourceOffset.X, Destination.Columns - DestOffset.X, MaximumSize.Width}.Min - 1
            For y = 0 To {Me.Rows - SourceOffset.Y, Destination.Rows - DestOffset.Y, MaximumSize.Height}.Min - 1
                Destination.Grid(DestOffset.X + x, DestOffset.Y + y) = Grid(SourceOffset.X + x, SourceOffset.Y + y)
            Next
        Next
    End Sub

    Private Sub ClearVisited()
        For x = 0 To Columns - 1
            For y = 0 To Rows - 1
                Grid(x, y).Visited = False
            Next
        Next
    End Sub
#Region "Overloads"
    Public Sub CopyTo(ByVal Destination As Maze, ByVal SourceOffset As Point, ByVal DestOffset As Point)
        CopyTo(Destination, SourceOffset, DestOffset, New Size(Integer.MaxValue, Integer.MaxValue))
    End Sub
    Public Sub CopyTo(ByVal Destination As Maze, ByVal MaximumSize As Size)
        CopyTo(Destination, Size.Empty, Size.Empty, MaximumSize)
    End Sub
    Public Sub CopyTo(ByVal Destination As Maze)
        CopyTo(Destination, Size.Empty, Size.Empty)
    End Sub
#End Region
End Class

Public Class Adjacent
    Property Direction As Sides
    Property Coords As Point
End Class